
''' A module for representing triangulations of surface bundles over the circle.

Provides one class: Bundle. '''

import flipper

class Bundle(object):
	''' This represents a triangulation of a surface bundle over the circle.
	
	It is specified by a triangulation of the surface, a triangulation of the
	bundle and an immersion map. Mapping classes can build their bundles and
	this is the standard way users are expected to create these.
	
	The following will be our standard examples:
	>>> import flipper
	>>> B4_1 = flipper.load('S_1_1').mapping_class('aB').bundle()
	>>> B8_21 = flipper.load('S_2_1').mapping_class('abcDF').bundle()
	>>> B4_1a = flipper.load('S_1_1').mapping_class('aB').bundle(canonical=False)
	>>> B8_21a = flipper.load('S_2_1').mapping_class('abcDF').bundle(canonical=False)
	'''
	def __init__(self, triangulation, triangulation3, immersion):
		assert(isinstance(triangulation, flipper.kernel.Triangulation))
		assert(isinstance(triangulation3, flipper.kernel.Triangulation3))
		assert(isinstance(immersion, dict))
		assert(all(triangle in immersion for triangle in triangulation))
		assert(all(isinstance(immersion[triangle], (list, tuple)) for triangle in triangulation))
		assert(all(len(immersion[triangle]) == 2 for triangle in triangulation))
		assert(all(isinstance(immersion[triangle][0], flipper.kernel.Tetrahedron) for triangle in triangulation))
		assert(all(isinstance(immersion[triangle][1], flipper.kernel.Permutation) for triangle in triangulation))
		
		self.triangulation = triangulation
		self.triangulation3 = triangulation3
		self.immersion = immersion
		# This is a dict mapping:
		#   triangle |---> (tetrahedra, perm).
		
		assert(self.triangulation3.is_closed())
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str(self.triangulation3)
	
	def snappy_string(self, name='flipper_triangulation', filled=True):
		''' Return the SnapPy string describing this triangulation.
		
		If filled=True then the fake cusps are filled along their fibre slope.
		
		>>> try:
		...     from snappy import Manifold  # SnapPy might not be installed.
		...     Manifold(B4_1.snappy_string()).is_isometric_to(Manifold('4_1'))
		...     Manifold(B8_21.snappy_string()).is_isometric_to(Manifold('8_21'))
		...     Manifold(B4_1a.snappy_string(filled=False)).is_isometric_to(Manifold('4_1'))
		...     Manifold(B8_21a.snappy_string(filled=False)).is_isometric_to(Manifold('8_21'))
		...     Manifold(B4_1.snappy_string(filled=False)).is_isometric_to(Manifold('4_1'))  # Could be False.
		...     Manifold(B8_21.snappy_string(filled=False)).is_isometric_to(Manifold('8_21'))  # Could be False.
		... except ImportError:
		...     True
		...     True
		...     True
		...     True
		...     True
		...     False
		True
		True
		True
		True
		True
		False
		'''
		
		fillings = [slope if filled_cusp else (0, 0) for filled_cusp, slope in zip(self.cusp_types(), self.fibre_slopes())] if filled else None
		return self.triangulation3.snappy_string(name, fillings)
	
	def __snappy__(self):
		return self.snappy_string()
	
	def cusp_types(self):
		''' Return the list of the type of each cusp.
		
		>>> B4_1.cusp_types(), sorted(B8_21.cusp_types())
		([False], [False, True])
		>>> B4_1a.cusp_types(), B8_21a.cusp_types()
		([False], [False])
		'''
		
		cusp_types = [None] * self.triangulation3.num_cusps
		
		for corner_class in self.triangulation.corner_classes:
			vertex = corner_class[0].vertex
			filled = vertex.filled
			for corner in corner_class:
				tetrahedron, permutation = self.immersion[corner.triangle]
				index = tetrahedron.cusp_indices[permutation(corner.side)]
				if cusp_types[index] is None:
					cusp_types[index] = filled
				else:
					assert(cusp_types[index] == filled)
		
		return cusp_types
		
	def fibre_slopes(self):
		''' Return the list of fibre slopes on each cusp. '''
		
		LONGITUDES, MERIDIANS = flipper.kernel.triangulation3.LONGITUDES, flipper.kernel.triangulation3.MERIDIANS
		
		slopes = [None] * self.triangulation3.num_cusps
		
		for index in range(self.triangulation3.num_cusps):
			meridian_intersection, longitude_intersection = 0, 0
			for corner_class in self.triangulation.corner_classes:
				corner = corner_class[0]
				tetra, perm = self.immersion[corner.triangle]
				if tetra.cusp_indices[perm(corner.side)] == index:
					for corner in corner_class:
						tetra, perm = self.immersion[corner.triangle]
						side, other = perm(corner.side), perm(3)
						meridian_intersection += tetra.peripheral_curves[MERIDIANS][side][other]
						longitude_intersection += tetra.peripheral_curves[LONGITUDES][side][other]
					slopes[index] = (longitude_intersection, -meridian_intersection)
					break
			else:
				raise RuntimeError('No vertex was mapped to this cusp.')
		
		return slopes
		
	def degeneracy_slopes(self):
		''' Return the list of degeneracy slopes on each cusp.
		
		This triangulation is must be veering. '''
		
		assert(self.triangulation3.is_veering())
		
		VEERING_LEFT, VEERING_RIGHT = flipper.kernel.triangulation3.VEERING_LEFT, flipper.kernel.triangulation3.VEERING_RIGHT
		TEMPS = flipper.kernel.triangulation3.TEMPS
		VERTICES_MEETING = flipper.kernel.triangulation3.VERTICES_MEETING
		EXIT_CUSP_LEFT, EXIT_CUSP_RIGHT = flipper.kernel.triangulation3.EXIT_CUSP_LEFT, flipper.kernel.triangulation3.EXIT_CUSP_RIGHT
		
		slopes = [None] * self.triangulation3.num_cusps
		
		cusp_pairing = self.triangulation3.cusp_identification_map()
		for index, cusp in enumerate(self.triangulation3.cusps):
			self.triangulation3.clear_temp_peripheral_structure()
			
			# Set the degeneracy curve into the TEMPS peripheral structure.
			# First find a good starting point:
			start_tetrahedron, start_side = cusp[0]
			edge_labels = [start_tetrahedron.get_edge_label(start_side, other) for other in VERTICES_MEETING[start_side]]
			for i in range(3):
				if edge_labels[(i+1) % 3] == VEERING_RIGHT and edge_labels[(i+2) % 3] == VEERING_LEFT:
					start_other = VERTICES_MEETING[start_side][i]
					break
			
			# Then walk around, never crossing through an edge where both ends veer the same way.
			current_tetrahedron, current_side, current_other = start_tetrahedron, start_side, start_other
			while True:
				current_tetrahedron.peripheral_curves[TEMPS][current_side][current_other] += 1
				if start_tetrahedron.get_edge_label(current_side, current_other) == VEERING_LEFT:
					leave = EXIT_CUSP_LEFT[(current_side, current_other)]
				else:
					leave = EXIT_CUSP_RIGHT[(current_side, current_other)]
				current_tetrahedron.peripheral_curves[TEMPS][current_side][leave] -= 1
				current_tetrahedron, current_side, current_other = cusp_pairing[(current_tetrahedron, current_side, leave)]
				if (current_tetrahedron, current_side, current_other) == (start_tetrahedron, start_side, start_other):
					break
			
			slopes[index] = self.triangulation3.slope()
		
		return slopes

def doctest_globs():
	''' Return the globals needed to run doctest on this module. '''
	
	B4_1 = flipper.load('S_1_1').mapping_class('aB').bundle()
	B8_21 = flipper.load('S_2_1').mapping_class('abcDF').bundle()
	B4_1a = flipper.load('S_1_1').mapping_class('aB').bundle(canonical=False)
	B8_21a = flipper.load('S_2_1').mapping_class('abcDF').bundle(canonical=False)
	
	return {'B4_1': B4_1, 'B8_21': B8_21, 'B4_1a': B4_1a, 'B8_21a': B8_21a}





''' Flipper is a program for computing the action of mapping classes on
laminations on punctured surfaces using ideal triangulation coordinates.

It can decide the Nielsen--Thurston type of a given mapping class and,
for pseudo-Anosov mapping classes, construct a layered, veering
triangulation of their mapping torus, as described by Agol.

Get started by starting the GUI:
	> import flipper.app
	> flipper.app.start()
or by creating a Triangulation using the helper function:
	> flipper.create_triangulation(...)
or from an isomorphism signature:
	> flipper.create_triangulation_from_iso_sig(...)
or by loading one of the provided EquippedTriangulations using:
	> flipper.load(...) '''

from flipper.version import __version__

# We'll only import the bare minimum. This way people missing packages
# can still use the flipper kernel at least.
# import flipper.application  # Uses tkinter.
import flipper.kernel
# import flipper.example  # Uses snappy.
import flipper.doc
# import flipper.test  # Uses snappy.
import flipper.profile

from flipper.load import load
from flipper.census import census

# Set up really short names for the most commonly used classes and functions by users.
create_triangulation = flipper.kernel.create_triangulation
triangulation_from_iso_sig = flipper.kernel.triangulation_from_iso_sig
norm = flipper.kernel.norm

StringType = flipper.kernel.StringType
IntegerType = flipper.kernel.IntegerType

AbortError = flipper.kernel.AbortError
ApproximationError = flipper.kernel.ApproximationError
AssumptionError = flipper.kernel.AssumptionError
ComputationError = flipper.kernel.ComputationError
FatalError = flipper.kernel.FatalError

# Finally load in the constants. These might require flipper to do the calculations
# so we can't create them before this point.
import flipper.kernel.constants





''' A module for interacting with various symbolic calculation libraries.

This module selects and imports the appropriate interface for manipulating algebraic numbers.
Currently there is an interface to sage, cypari and a dummy library. A library should raise an
ImportError on import if it cannot be used. The dummy library can always be used, even in
Python 3.

There used to be an interface to SymPy but this turned out to be extremely unreliable. There
were 9x9 invertible matrices for which it could only find 1 eigenvalue!

The dummy library also includes a basic implementation of the gram-schmidt orthonormalisation
process and project function that other libraries may be able to make use of.

Each library provides one function:
	directed_eigenvector(action_matrix, condition_matrix, vector)

More interfaces can be added here. '''

from importlib import import_module

# Add new libraries here in order.
INTERFACES = ['sage', 'cypari', 'dummy']

def load_interface():
	''' Return the first available interface.
	
	If none are available then an ImportError will be raised.
	This should nver happen thanks to the dummy interface provided. '''
	
	for interface in INTERFACES:
		try:
			module = import_module('flipper.kernel.interface.' + interface)
			# We could add some code here to find out which interface was loaded.
			return module
		except ImportError:
			pass
	
	raise ImportError('No symbolic computation interface available.')

def directed_eigenvector(action_matrix, condition_matrix):
	''' Return an interesting eigenvector of action_matrix which lives inside of the cone C, defined by condition_matrix.
	
	An eigenvector is interesting if its corresponding eigenvalue is: real, > 1, irrational and bigger than all
	of its Galois conjugates.
	
	Raises a ComputationError if it cannot find an interesting vectors in C.
	Assumes that C contains at most one interesting eigenvector. '''
	
	return load_interface().directed_eigenvector(action_matrix, condition_matrix)

